# coding=utf-8
import os
import time
from typing import List

import dbf
from datetime import datetime

import random
from common import Order,CancelOrder,RawFunc

showSQL = False

table_place_order = None
table_cancel_order = None
table_raw_func = None

def _create_table(filename:str, fields:List[str]):
    if filename.endswith(".dbf") is False:
        filename=filename+".dbf"
    table = dbf.Table(
            filename=filename,
            field_specs="; ".join(fields),
            on_disk=True,
            # default_data_types='enhanced',
            codepage='utf8',  # 默认为ASCII，这里需要指定为utf8编码，否则数据读取会出现乱码，且写入中文会报错
        )
    table.open(mode=dbf.READ_WRITE)
    return table


def create_place_order_table(filename: str):
    # 定义报单字段
    fields = [
        "sid C(64)",
        "account_id C(64)",
        "symbol C(32)",
        "volume N(16,0)",
        "order_type N(2,0)",
        "order_biz N(8,0)",
        "price N(16,2)",
        "comment C(255)",
    ]
    return _create_table(filename, fields)


def create_cance_order_table(filename: str):
    # 定义撤单字段
    fields = [
        "sid C(64)",
        "comment C(255)",
    ]
    return _create_table(filename, fields)


def create_func_table(filename: str):
    # 定义功能号字段
    fields = [
        "rid C(64)",
        "account_id C(64)",
        "func_id C(64)",
        "func_params C(255)",
        "comment C(255)",
    ]
    return _create_table(filename, fields)


def _generate_dbf(folder:str, typ:str)->str:
    return os.path.join(folder,"{}.{}.dbf".format(datetime.now().strftime('%Y%m%d%H%M%S.%f'), typ))


def generate_place_order_dbf(folder:str)->str:
    return _generate_dbf(folder, "order")


def generate_cancel_order_dbf(folder:str)->str:
    return _generate_dbf(folder,"cancel_order")


def generate_func_dbf(folder:str)->str:
    return _generate_dbf(folder, "func_request")


def query(dbf_file: str, condition: str)->List:
    '''
    使用SQL查询数据
    '''
    res = []
    table = dbf.Table(dbf_file, codepage='utf8')
    table.open()
    fields = table.field_names
    sql = f'select * from "{dbf_file}" where {condition}'
    if showSQL:
        print("\n【SQL】", sql)
    rows = table.query(sql)
    if rows is not None and len(rows)>0:
        for row in rows:
            data = {}
            for field in fields:
                val = row[field]
                if isinstance(val, str):
                    val=str(val).strip()
                data[field]=val
            res.append(data)
    table.close()
    return res


def place_order(dbf_dir: str, order: Order):
    """
    单笔委托
    """
    return place_orders(dbf_dir, [order])[0]


def place_orders(dbf_dir: str, orders: List[Order])->List[str]:
    """
    批量委托
    """
    global table_place_order
    sids = []
    if table_place_order is None:
        filename=generate_place_order_dbf(dbf_dir)
        table_place_order = create_place_order_table(filename)

    for order in orders:
        ord = (order.sid, order.account_id, order.symbol, order.volume, order.order_type, order.order_biz, order.price, order.comment)
        table_place_order.append(ord)
        sids.append(order.sid)

    table_place_order._table.flush()

    return sids


def cancel_order(dbf_dir:str, order:CancelOrder)->str:
    return cancel_orders(dbf_dir, [order])


def cancel_orders(dbf_dir:str, orders:List[CancelOrder])->List[str]:
    """
    批量撤单
    """
    global table_cancel_order
    sids = []
    if table_cancel_order is None:
        filename=generate_cancel_order_dbf(dbf_dir)
        table_cancel_order = create_cance_order_table(filename)
    for order in orders:
        ord = (order.sid, order.comment)
        table_cancel_order.append(ord)
        sids.append(order.sid)

    table_cancel_order._table.flush()

    return sids


def raw_func(dbf_dir:str, order: RawFunc)->str:
    return raw_funcs(dbf_dir, [order])[0]


def raw_funcs(dbf_dir:str, orders:List[RawFunc])->List[str]:
    """
    批量撤单
    """
    global table_raw_func
    rids = []
    if table_raw_func is None:
        filename=generate_func_dbf(dbf_dir)
        table_raw_func = create_func_table(filename)

    for order in orders:
        ord = (order.rid, order.account_id, order.func_id, order.func_params, order.comment)
        table_raw_func.append(ord)
        rids.append(order.rid)
    
    table_raw_func._table.flush()

    return rids


def get_position(out_path: str, account_id: str)->List:
    dbf_file = os.path.join(out_path,f'{account_id}/position.dbf')
    return query(dbf_file, f'account_id.strip() == "{account_id}"')


def get_cash(out_path: str, account_id: str)->List:
    dbf_file = os.path.join(out_path,f'{account_id}/cash.dbf')
    return query(dbf_file, f'account_id.strip() == "{account_id}"')


def get_orders(out_path: str, account_id: str, sids: List[str] = [], status: List[int] = [])->List:
    dbf_file = os.path.join(out_path,f'{account_id}/order_status.dbf')
    condition = [f'account_id.strip() == "{account_id}"']
    if len(sids)>0:
        condition.append(f'sid.strip() in {sids}')
    
    if len(status)>0:
        condition.append(f'status in {status}')

    return query(dbf_file, " and ".join(condition))


def get_orders_changelog(out_path: str, account_id: str, sids: List[str] = [], status: List[int] = [])->List:
    dbf_file = os.path.join(out_path, f'{account_id}/order_status_change.dbf')
    condition = [f'account_id.strip() == "{account_id}"']
    if len(sids)>0:
        condition.append(f'sid.strip() in {sids}')
    
    if len(status)>0:
        condition.append(f'status in {status}')

    return query(dbf_file, " and ".join(condition))


def get_execrpt(out_path: str, account_id: str, sids: List[str] = [])->List:
    dbf_file = os.path.join(out_path, f'{account_id}/execution_report.dbf')
    condition = [f'account_id.strip() == "{account_id}"']
    if len(sids)>0:
        condition.append(f'sid.strip() in {sids}')

    return query(dbf_file, " and ".join(condition))


def get_raw_func(out_path: str, account_id: str, rids: List[str] = [])->List:
    dbf_file = os.path.join(out_path, f'{account_id}/func_response.dbf')
    condition = [f'account_id.strip() == "{account_id}"']
    if len(rids)>0:
        condition.append(f'rid.strip() in {rids}')

    return query(dbf_file, " and ".join(condition))


def finish_place_order():
    global table_place_order
    table_place_order.close()
    table_place_order = None


def finish_cancel_order():
    global table_cancel_order
    table_cancel_order.close()
    table_cancel_order = None


def finish_raw_func():
    global table_raw_func
    table_raw_func.close()
    table_raw_func = None


if __name__=="__main__":
    '''
    【温馨提示】
    1. 当前示例程序仅供参考，请勿用于实盘交易
    2. 打开dbf文件时，请用codepage参数指定字符编码为utf8，以免读写文件出现中文乱码报错问题
    3. 由于`dbf`文件规范本身字段值长度(254个字符长度)的限制，`功能号`扫单不适合使用`dbf`文件单进行扫单，建议使用`csv`
    '''
    symbols = [
        "SHSE.600000", "SHSE.600004", "SHSE.600006", "SHSE.600007", "SHSE.600008", "SHSE.600009", "SHSE.600010",
        "SHSE.600012", "SHSE.600015", "SHSE.600016", "SHSE.600017", "SHSE.600018", "SHSE.600019", "SHSE.600020",
        "SHSE.600022", "SHSE.600023", "SHSE.600025", "SHSE.600026", "SHSE.600027", "SHSE.600028", "SHSE.600029",
        "SHSE.600031", "SHSE.600033", "SHSE.600035", "SHSE.600036", "SHSE.600037", "SHSE.600038", "SHSE.600039",
        "SHSE.600050", "SHSE.600051", "SHSE.600052", "SHSE.600053", "SHSE.600054", "SHSE.600055", "SHSE.600056",
        "SHSE.600058", "SHSE.600059", "SHSE.600060", "SHSE.600061", "SHSE.600062", "SHSE.600063", "SHSE.600064",
        "SHSE.600067", "SHSE.600068", "SHSE.600069", "SHSE.600070", "SHSE.600071", "SHSE.600072", "SHSE.600073",
        "SHSE.600076", "SHSE.600077", "SHSE.600078", "SHSE.600079", "SHSE.600080", "SHSE.600081", "SHSE.600082",
        "SHSE.600084", "SHSE.600085", "SHSE.600086", "SHSE.600088", "SHSE.600089", "SHSE.600090", "SHSE.600093",
        "SHSE.600095", "SHSE.600096", "SHSE.600097", "SHSE.600098", "SHSE.600099", "SHSE.600100", "SHSE.600101",
        "SHSE.600104", "SHSE.600105", "SHSE.600106", "SHSE.600107", "SHSE.600108", "SHSE.600109", "SHSE.600110",
        "SHSE.600112", "SHSE.600113", "SHSE.600114", "SHSE.600115", "SHSE.600116", "SHSE.600117", "SHSE.600118",
        "SHSE.600120", "SHSE.600121", "SHSE.600122", "SHSE.600123", "SHSE.600125", "SHSE.600126", "SHSE.600127",
        "SHSE.600129", "SHSE.600130", "SHSE.600131", "SHSE.600132", "SHSE.600133", "SHSE.600135", "SHSE.600136",
    ]

    # 文件单结果输出路径：登录掘金3终端 -> 手工交易 -> 文件单 -> 文件单输入 -> 新增扫单的扫单路径 
    out_path = r'D:\workspace\be\sanner-strategy\examples\out'
    
    # 文件单结果输入路径：登录掘金3终端 -> 手工交易 -> 文件单 -> 文件单输出-> 输出路径
    in_path = r'D:\workspace\be\sanner-strategy\examples\in'

    # 仿真/实盘的账户ID
    account_id = '0a0d76c0-32ba-11eb-9210-001018b67b94'

    order = Order(
            account_id=account_id,
            symbol=random.choice(symbols), # 随机选择一个标的进行报单操作
            volume=100,
            order_type=2, # 市价单
            order_biz=1,
            price=0,
            comment='我的委托'
        )

    showSQL = True

    # 报单
    sid = place_order(dbf_dir=in_path, order=order)
    print(sid, order.symbol)

    # 撤单
    cancel_order(dbf_dir=in_path, order=CancelOrder(sid=sid))

    # 功能号。由于dbf本身规范的限制，dbf不适合用来做文件单扫单，建议使用csv
    # raw_func(dbf_dir=in_path, order=RawFunc(account_id=account_id,func_id='hello', func_params='{"param1":"value1"}'))

    time.sleep(1)

    start = datetime.now()
    cash = get_cash(out_path, account_id)
    print("\n【资金】", datetime.now() - start, len(cash), cash)

    start = datetime.now()
    pos = get_position(out_path, account_id)
    print("\n【持仓】", datetime.now() - start, len(pos), pos)
    
    start = datetime.now()
    orders = get_orders(out_path, account_id, sids=[sid])
    print("\n【委托】", datetime.now() - start, len(orders), orders)

    start = datetime.now()
    orders_changelog = get_orders_changelog(out_path, account_id, sids=[sid], status=[1, 3, 10])
    print("\n【委托流水】", datetime.now() - start, len(orders_changelog), orders_changelog)

    start = datetime.now()
    raw_func = get_raw_func(out_path, account_id, rids=[sid])
    print("\n【功能号】", datetime.now() - start, len(raw_func), raw_func)

    start = datetime.now()
    execrpts = get_execrpt(out_path, account_id, sids=[sid])
    print("\n【委托回报】", datetime.now() - start, len(execrpts), execrpts)